// ***************************************************************
// LZ.H				CLIB I/O datafile interceptor			  v2.0
//
// (C) Copyright 1997-2000 7Gods demo group http://7gods.sk
// 
// ***************************************************************
#ifndef __LZ_H
#define __LZ_H

#include "types.h"

// ***************************************************************
// Error codes
#define DTF_OK				0
#define DTF_FILENOTFOUND	1
#define DTF_BADFILE			2
#define DTF_NOMEMORY		3

// ***************************************************************
#define BLOCK_LEN 32768			// compression algorithm block size
#define SIZE2BLOCKS(size) (((size) + BLOCK_LEN - 1) / BLOCK_LEN)
#define OFFSET2BLOCK(off) ((off) / BLOCK_LEN)

// ***************************************************************
// block of file on disk
typedef struct {
	DWORD offset;				// offset in datafile of the given block
    DWORD cksum;				// ADLER-32 checksum
    WORD length;				// length of the compressed block (uncompressed is almost always 32k)
} CBLOCK;

// ***************************************************************
// File in memory buffer
typedef struct _ST_FILE {
    FILE *dataf;				// datafile file handle
    long  dtfOffset;            // datafile offset in the datafile file
    DWORD fp;					// file pointer
    DWORD length;               // uncompressed length
    WORD nblocks;               // number of blocks
    WORD curBlock;				// index of currently loaded block = -1
    WORD prevBlock;             // index of currently loaded block = -1
    WORD cbLen;                 // length of the current block
    CBLOCK *blocks;				// array of compressed blocks
    BYTE *cbData;				// currently uncompressed block

    _ST_FILE() { blocks = 0; fp = 0; length = 0; cbData = 0; };
    ~_ST_FILE() { delete [] blocks; delete [] cbData; };

} ST_FILE;

// ***************************************************************
// File on disk
typedef struct _ST_STOREDFILE {
    char *name;			  		// virtual path
    FILE *dataf;				// FILE pointer to the file's datafile
    long  dtfOffset;            // datafile offset in the datafile file
    DWORD length;               // uncompressed length
    DWORD offset;               // file's offset in its datafile

    _ST_STOREDFILE() { name = 0; };
//    ~_ST_STOREDFILE() { delete [] name; };    FAKE! viz open_datafile

} ST_STOREDFILE;

/*****************************************************************
// Datafile format

OFFSET     LEN     DESC

  0         4      datafile creation time (time_t structure)
  4         4      number of files
  8         4      offset to file table
 12         4      ADLER-32 checksum of previous 12 bytes

 16         ?      file blocks (CBLOCK) 

..
..
..

FILE TABLE

  0         2      dlzka cesty ku suboru (aj s ukoncovacou 0)
  2         ?      relativna cesta ku suboru (ASCIIZ)
  ?         4	   povodna dlzka suboru
  ?         4      offset v datafile

*****************************************************************/


// ***************************************************************
// Scramble constant (primitive encryption)
#define XOR_BUF_CONST 0x732C2E17

// ***************************************************************
// Every file block gets XORed after compression
inline xorBuf(void *buffer,int len,DWORD init) {

    len >>= 2;				// len DWORD xor
    DWORD *buf = (DWORD *)buffer;
    if (len >= 4)
		do {
			buf[0] ^= init; init += XOR_BUF_CONST;
			buf[1] ^= init; init += XOR_BUF_CONST;
			buf[2] ^= init; init += XOR_BUF_CONST;
			buf[3] ^= init; init += XOR_BUF_CONST;
            len -= 4;
            buf += 4;
	    } while (len >= 4);

	BYTE *bufByte = (BYTE *)buf;
    if (len)
    	do {
			*bufByte++ ^= (BYTE)init;
    	    init += XOR_BUF_CONST;
		} while (--len > 0);
}

#endif
