// ***************************************************************
// LZIO.CPP			CLIB I/O datafile interceptor			  v2.0
//
// (C) Copyright 1997-2000 7Gods demo group http://7gods.sk
// 
// supported function calls: 
// fopen, fclose, fread, ftell, fgets, fseek, fwrite, fcloseall
// ***************************************************************

#ifdef WIN32
    #include <windows.h>
    CRITICAL_SECTION    csDatafileRequest;
#else
    #define EnterCriticalSection(X)
    #define LeaveCriticalSection(X)
    #define InitializeCriticalSection(X)
    #define DeleteCriticalSection(X)
#endif

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys\stat.h>
#include "lzo1x.h"
#include "lz.h"

//#define LOGFILE
#define MAX_OPEN_FILES 32		// maximum number of simultaneously opened datafile files
#define MAX_DATAFILES  16		// maximum number of simultaneously open datafiles

#undef EXTERN_C
#define EXTERN_C extern "C"

// _CLIB_VERSION was used only with Watcom C++ so this is useless now
#ifdef _CLIB_VERSION

#define CLIB_FOPEN  	 x6q07
#define CLIB_FCLOSE 	 x9c66a
#define CLIB_FREAD  	 x0p9w
#define CLIB_FTELL  	 x7hi2
#define CLIB_FGETS  	 x1ll8
//#define CLIB_FSCANF      x6ep9
#define CLIB_FSEEK       x9cv3
#define CLIB_FWRITE 	 x3y6zb
#define CLIB_FCLOSEALL   x83bf731n

#define OCC_FOPEN  	 fopen
#define OCC_FCLOSE 	 fclose
#define OCC_FREAD  	 fread
#define OCC_FTELL  	 ftell
#define OCC_FGETS  	 fgets
//#define OCC_FSCANF   fscanf
#define OCC_FSEEK    fseek
#define OCC_FWRITE 	 fwrite
#define OCC_FCLOSEALL   fcloseall

extern "C" {
	FILE * CLIB_FOPEN(const char *_filename,const char *_mode);
	int    CLIB_FCLOSE(FILE *_fp);
	size_t CLIB_FWRITE(const void *_ptr,size_t _size,size_t _n,FILE *_fp);
	size_t CLIB_FREAD(void *_ptr,size_t _size,size_t _n,FILE *_fp);
	int    CLIB_FSEEK(FILE *_fp,long int _offset,int _whence);
	char * CLIB_FGETS(char *s, int n, FILE *stream);
//    int    CLIB_FSCANF(FILE *stream, const char *format, ...);
	long   CLIB_FTELL(FILE *stream);
	int    CLIB_FCLOSEALL(void);
};

#else

#define CLIB_FOPEN  	 fopen
#define CLIB_FCLOSE 	 fclose
#define CLIB_FREAD  	 fread
#define CLIB_FTELL  	 ftell
#define CLIB_FGETS  	 fgets
//#define CLIB_FSCANF      fscanf
#define CLIB_FSEEK       fseek
#define CLIB_FWRITE 	 fwrite
#define CLIB_FCLOSEALL   fcloseall

#define OCC_FOPEN  	 _lz_fopen
#define OCC_FCLOSE 	 _lz_fclose
#define OCC_FREAD  	 _lz_fread
#define OCC_FTELL  	 _lz_ftell
#define OCC_FGETS  	 _lz_fgets
//#define OCC_FSCANF   _lz_fscanf
#define OCC_FSEEK    _lz_fseek
#define OCC_FWRITE 	 _lz_fwrite
#define OCC_FCLOSEALL   _lz_fcloseall

#endif

char *lz_io_copyright = "\n\n"
	"-----------------------------------------------\n"
	"LZIO (OCC) Module. Packed datafile I/O interceptor for C/C++\n\n"
    "Copyright (C) 1997-2000 by 7Gods.\n"
    "All rights reserved.\n"
	"-----------------------------------------------\n\n";

#include "logfile.cc"

// ***************************************************************
// static global variables
static FILE *datafiles[MAX_DATAFILES] = { 0, };		// array of opened datafiles
static int ndatafiles = 0;							// # of opened datafiles
static ST_STOREDFILE *stfile = 0;					// array of files in opened datafiles
static int nfiles = 0;								// # of files in opened datafiles
static ST_FILE ofile[MAX_OPEN_FILES];				// array of opened files
static int freefile = 0;							// index in OFILE to a free member

int _init_error(int error);
int readcurblock(ST_FILE *stf);

// ***************************************************************
// ADLER-32 CHECKSUM
// adler32 checksum Mark Adler <madler@alumni.caltech.edu>
// http://quest.jpl.nasa.gov/zlib

#define LZO_BASE 65521u /* largest prime smaller than 65536 */
#define LZO_NMAX 5552
/* NMAX is the largest n such that 255n(n+1)/2 + (n+1)(BASE-1) <= 2^32-1 */

#define LZO_DO1(buf,i)  {s1 += buf[i]; s2 += s1;}
#define LZO_DO2(buf,i)  LZO_DO1(buf,i); LZO_DO1(buf,i+1);
#define LZO_DO4(buf,i)  LZO_DO2(buf,i); LZO_DO2(buf,i+2);
#define LZO_DO8(buf,i)  LZO_DO4(buf,i); LZO_DO4(buf,i+4);
#define LZO_DO16(buf,i) LZO_DO8(buf,i); LZO_DO8(buf,i+8);

static DWORD adler32(DWORD adler, const BYTE *buf, unsigned len) {

	DWORD s1 = adler & 0xffff;
	DWORD s2 = (adler >> 16) & 0xffff;
	int k;

	if (buf == 0) return 1;

	while (len > 0) {
		k = len < LZO_NMAX ? (int) len : LZO_NMAX;
		len -= k;
		if (k >= 16)
			do {
				LZO_DO16(buf,0);
				buf += 16;
				k -= 16;
			} while (k >= 16);
		if (k != 0)
			do {
				s1 += *buf++;
				s2 += s1;
			} while (--k > 0);
		s1 %= LZO_BASE;
		s2 %= LZO_BASE;
	}
	return (s2 << 16) | s1;
}

// *********************************************************************
// fgets
EXTERN_C char *OCC_FGETS(char *s, int n, FILE *stream) {

    DWORD fidx = ((DWORD)stream)-1;
    if (stfile == 0 || fidx >= MAX_OPEN_FILES) return CLIB_FGETS(s,n,stream);

    ST_FILE *stf = &ofile[fidx];
    if (readcurblock(stf)) return 0;

    int bfp = stf->fp % BLOCK_LEN;
	int reminder = stf->cbLen - bfp;
    int pos = 0;

	LOG3("Starting FGETS for %d characters on file %d\n",n,fidx);
    while (n > 0) {
    	if (reminder <= 0) {
        	// potrebujem dalsi blok
            // ak bol toto posledny blok, tak smola => koniec suboru
    		if (stf->curBlock == stf->nblocks - 1) break;
			if (readcurblock(stf)) return 0; //break;
            reminder = stf->cbLen;
            bfp = 0;
        }
    	size_t reallen = reminder >= n ? n : reminder;

        // kopirujem dovtedy, pokial nie je newline
        char *from = (char *)(stf->cbData + bfp);
        for (unsigned i = 0; i < reallen; i++,n--,reminder--) {
    	    if (from[i] != 0x0D) s[pos++] = from[i];
            stf->fp++;
    	    if (from[i] == 0x0A) {
				s[pos++] = 0;
				LOG3("READ %d characters: \"%s\"",pos,s);
                return s;
			}
        }
    }

    if (pos == 0) return 0;
	s[pos++] = 0;
	LOG3("READ %d characters: \"%s\"",pos,s);
	return s;
}

// *********************************************************************
// ftell
EXTERN_C long OCC_FTELL(FILE *stream) {

    DWORD fidx = ((DWORD)stream)-1;
    if (stfile == 0 || fidx >= MAX_OPEN_FILES) return CLIB_FTELL(stream);

    return ofile[fidx].fp;
}

// *********************************************************************
// fseek
EXTERN_C int OCC_FSEEK(FILE *_fp,long int _offset,int _whence) {

    DWORD fidx = ((DWORD)_fp)-1;
    if (stfile == 0 || fidx >= MAX_OPEN_FILES) return CLIB_FSEEK(_fp,_offset,_whence);

    ST_FILE *stf = &ofile[fidx];

    LOG2("Fseek invoked on file %d\n",fidx);
    switch (_whence) {
    case SEEK_SET:
		if (_offset < 0 || (unsigned long)_offset > stf->length) return 1;
		stf->fp = _offset;
	    LOG2("Fseek SET ok, new pointer at %d\n",stf->fp);
		break;
    case SEEK_CUR:
		if (stf->fp + _offset > stf->length) return 1;
		stf->fp += _offset;
        break;
    case SEEK_END:
		if (_offset > 0 || (unsigned long)-_offset > stf->length) return 1;
		stf->fp = stf->length + _offset;
	    LOG2("Fseek END ok, new pointer at %d\n",stf->fp);
        break;
    default:
    	return 2;
    };
	return 0;
}

// *********************************************************************
// statx
EXTERN_C int statx(char *path, struct stat *statbuf) {

	if (stfile == 0) return stat(path,statbuf);

    char temp[_MAX_PATH];
    int i;

    strcpy(temp,path); strupr(temp);
    for (i=0; i<nfiles; i++)
		if (strcmp(stfile[i].name,temp)==0) break;

	if (i == nfiles) return stat(path,statbuf);

    statbuf->st_nlink = 1;
    statbuf->st_size = stfile[i].length;

    return 0;
}

// *********************************************************************
// readcurblock (CORE function)
// insures that cbData catains block that file pointer points into
static int readcurblock(ST_FILE *stf) {

    WORD block = (WORD) OFFSET2BLOCK(stf->fp);
    if ( block != stf->curBlock) {
        LOG3("Start readcurclock for pointer %d, block %u\n",stf->fp,block);
        stf->curBlock = block;
        EnterCriticalSection(&csDatafileRequest);
        if (CLIB_FSEEK(stf->dataf,stf->blocks[block].offset + stf->dtfOffset,SEEK_SET) != 0) {
            LeaveCriticalSection(&csDatafileRequest);
            return 1;
        }
        LOG3("Read block from %d length %u\n",stf->blocks[block].offset,
			stf->blocks[block].length);
        unsigned blen = stf->blocks[block].length;
        if (blen > 0) {
            BYTE tmpBuffer[BLOCK_LEN];
            LOG2("Clib read HEHE to %p\n",tmpBuffer);
            if (CLIB_FREAD(tmpBuffer,blen,1,stf->dataf) != 1) {
                LOG("Clib read FAILED\n");
                LeaveCriticalSection(&csDatafileRequest);
                return 1;
            }
            LOG("block read ok\n");
            LeaveCriticalSection(&csDatafileRequest);
            xorBuf(tmpBuffer,blen,0);
            LOG("block xored ok\n");
            unsigned tmplen = BLOCK_LEN;
            if (lzo1x_decompress_asm_fast(tmpBuffer,blen,
                stf->cbData,&tmplen,0) != LZO_E_OK) return 2;
            stf->cbLen = (WORD)tmplen;
            LOG("block decompressed ok\n");
        }
        else {
	        LOG2("Read stored from %d\n",stf->blocks[block].offset);
            if (CLIB_FREAD(stf->cbData,BLOCK_LEN,1,stf->dataf) != 1) {
                LeaveCriticalSection(&csDatafileRequest);
                return 1;
            }
            LeaveCriticalSection(&csDatafileRequest);
            xorBuf(stf->cbData,BLOCK_LEN,0);
            stf->cbLen = BLOCK_LEN;
        }
    }
    return 0;
}

// *********************************************************************
// fread
EXTERN_C size_t OCC_FREAD(void *_ptr,size_t _size,size_t _n,FILE *_fp) {

    DWORD fidx = ((DWORD)_fp)-1;
    if (stfile == 0 || fidx >= MAX_OPEN_FILES) return CLIB_FREAD(_ptr,_size,_n,_fp);

    ST_FILE *stf = &ofile[fidx];
    if (stf->fp >= stf->length || readcurblock(stf)) return 0;

    size_t len = _size * _n;
    if (len <= 0) return 0;
    BYTE *tmpPtr = (BYTE *)_ptr;
    int bfp = stf->fp % BLOCK_LEN;
	int reminder = stf->cbLen - bfp;

    while (len > 0) {
        LOG4("FREAD: toRead %d, block = %d, reminder %d\n",len,stf->curBlock,reminder);
        if (reminder <= 0) {
        	// potrebujem dalsi blok
            // ak bol toto posledny blok, tak smola => koniec suboru
            LOG3("FREAD: curBock = %d, nblocks = %d\n",stf->curBlock,stf->nblocks);
            if (stf->curBlock == stf->nblocks - 1) return (_size * _n - len) / _size;
            LOG3("FREAD: %d != %d\n",stf->curBlock,stf->nblocks-1);
            if (readcurblock(stf)) return 0; //break;
            reminder = stf->cbLen;
            bfp = 0;
        }
    	size_t reallen = (size_t)reminder >= len ? len : reminder;
		memcpy(tmpPtr,stf->cbData + bfp,reallen);
        stf->fp += reallen;
        if ( (len -= reallen) == 0) break;
        bfp += reallen;
        tmpPtr += reallen;
        reminder -= reallen;
    }

    return (_size * _n - len) / _size;
}

// *********************************************************************
// fclose
// for datafile files always returns 0
EXTERN_C int OCC_FCLOSE(FILE *_fp) {

    DWORD fidx = ((DWORD)_fp)-1;
    if (stfile == 0 || fidx >= MAX_OPEN_FILES) return CLIB_FCLOSE(_fp);

	LOG2("Closing file %d ... ",fidx);
    // uvolnim strukturu STFILE
    delete [] ofile[fidx].blocks; ofile[fidx].blocks = 0;
    delete [] ofile[fidx].cbData; ofile[fidx].cbData = 0;

	LOG("closed\n");
    return 0;
}

// *********************************************************************
// fcloseall
EXTERN_C int OCC_FCLOSEALL(void) {

    if (stfile != 0)
		for (int i = 0; i<MAX_OPEN_FILES; i++) OCC_FCLOSE((FILE *)(i+1));
    return 0;	// ! nemozem volat fcloseall, lebo si zavriem datafiles !
}

// *********************************************************************
// fopen
// for reading with "R" "RB" returns fake datafile index as file handle
// for writing "W" "WB" and reading with other flags returns DOS file handle
EXTERN_C FILE *OCC_FOPEN(const char *_filename,const char *_mode) {

    char temp[_MAX_PATH];
    DWORD i;

    LOG3("Start fopen on %s, %s\n",_filename,_mode);
    if (stfile == 0 || strpbrk(_mode,"rR") == 0) return CLIB_FOPEN(_filename,_mode);

	LOG2("Opening file %s\n",_filename);

    strcpy(temp,_filename); strupr(temp);
    for (i = 0; i < strlen(temp); i++)
		if (temp[i] == '/') temp[i] = '\\';

    for (i = 0; i < (unsigned)nfiles; i++)
		if (stricmp(stfile[i].name,temp)==0) break;

    if (i == (unsigned)nfiles) return CLIB_FOPEN(_filename,_mode);
    if (freefile == MAX_OPEN_FILES) return 0;

    EnterCriticalSection(&csDatafileRequest);

	LOG2("File %s found.\n",_filename);
    // set offset in datafile
    if (CLIB_FSEEK(stfile[i].dataf,stfile[i].offset,SEEK_SET) !=0 ) return 0;

    int cf = freefile;
    // ofile structure initialization
    ofile[cf].dataf = stfile[i].dataf;
    ofile[cf].length = stfile[i].length;
    ofile[cf].fp = 0;
    ofile[cf].curBlock = 0xFFFF;

    ofile[cf].prevBlock = 0x0FFF;

    ofile[cf].nblocks = (WORD) SIZE2BLOCKS(stfile[i].length);
    ofile[cf].blocks = new CBLOCK [ofile[cf].nblocks];
    ofile[cf].cbData = new BYTE [BLOCK_LEN + 4];	// for FAST decompressor

    ofile[cf].dtfOffset = stfile[i].dtfOffset;

    LOG3("File has %u blocks (length = %u).\n",ofile[cf].nblocks,stfile[i].length);
    // reading blocks
    if (CLIB_FSEEK(stfile[i].dataf,stfile[i].offset,SEEK_SET)) return 0;
	if (CLIB_FREAD( ofile[cf].blocks, sizeof(CBLOCK), ofile[cf].nblocks, ofile[cf].dataf)
		!= ofile[cf].nblocks) return 0;

    // update freefile
    for (freefile = 0;
		freefile < MAX_OPEN_FILES && ofile[freefile].cbData != 0;
		freefile++);

	LOG3("Open OK, assigned %d, next freefile %d\n",cf,freefile);
    LeaveCriticalSection(&csDatafileRequest);
    return (FILE *)(cf+1);
}

// *********************************************************************
// fwrite
// there is NO support for writing into datafile in this release
EXTERN_C size_t OCC_FWRITE(const void *_ptr,size_t _size,size_t _n,FILE *_fp) {

    DWORD fidx = ((DWORD)_fp)-1;

    if (stfile == 0 || fidx >= (unsigned)nfiles) return CLIB_FWRITE(_ptr,_size,_n,_fp);
    else return 0;
}

// *********************************************************************
// open_datafile
// prepares a datafile for reading
extern "C" int open_datafile(const char *_filename,long soffset) {

    DWORD header[4];
    ST_STOREDFILE fdesc,*stf;
    DWORD taboffs;
    FILE *dtf;
    int n;

    if (soffset == -1) soffset = 0;

    if ( (dtf = CLIB_FOPEN(_filename,"rb")) == NULL) return DTF_FILENOTFOUND;
    if (ndatafiles >= MAX_DATAFILES - 1) return DTF_FILENOTFOUND;
	datafiles[ndatafiles++] = dtf;

    if (CLIB_FSEEK(dtf,soffset,SEEK_SET) != 0) return DTF_BADFILE;

    if (CLIB_FREAD(header,4,4,dtf) != 4) return DTF_BADFILE;
    if (header[3] != adler32(0,(BYTE *)header,12)) return DTF_BADFILE;
    n = header[1]; taboffs = header[2];

    if ( (stf = new ST_STOREDFILE[n]) == 0) return DTF_NOMEMORY;

    if (CLIB_FSEEK(dtf,taboffs + soffset,SEEK_SET)) return DTF_BADFILE;
    DWORD initxor = taboffs;
	LOG2("Datafile %s opened OK\n",_filename);
    for (int i = 0; i < n; i++) {
		WORD namelen;
		if (CLIB_FREAD(&namelen,2,1,dtf)!=1) return DTF_BADFILE;
		stf[i].name = new char[namelen];
		if (CLIB_FREAD(stf[i].name,namelen,1,dtf)!=1) return DTF_BADFILE;
        xorBuf(stf[i].name,namelen,initxor);
		if (CLIB_FREAD(&stf[i].length,8,1,dtf)!=1) return DTF_BADFILE;
        xorBuf(&stf[i].length,8,initxor);
        stf[i].dataf = dtf;
        stf[i].offset += soffset;
        stf[i].dtfOffset = soffset;
        initxor *= 0x17BC3;				// another magic number
    }

    if (stfile == 0) {	// the first time
    	nfiles = n;
	    stfile = stf;
		if (lzo_init() != LZO_E_OK) return 666;
		LOG2("--- Initialization passed (%d files)\n",nfiles);
    }
    else {
		ST_STOREDFILE *oldstfile = stfile;
	    if ( (stfile = new ST_STOREDFILE[nfiles + n]) == 0) return DTF_NOMEMORY;
        memcpy(stfile,oldstfile,nfiles*sizeof(ST_STOREDFILE));
        memcpy(stfile + nfiles,stf,n*sizeof(ST_STOREDFILE));
        delete [] stf;
		LOG2("%d new files added\n",n);
    	nfiles += n;
    }

	LOG("Current files:\n");
	for (int s = 0; s < nfiles; s++) LOG2("%s\n",stfile[s].name);

    InitializeCriticalSection(&csDatafileRequest);
    return DTF_OK;
};

// *********************************************************************
// close_datafile
EXTERN_C void close_datafile(void) {

   	int i;
    if (stfile != 0) {
		for (i = 0; i<MAX_OPEN_FILES; i++) OCC_FCLOSE((FILE *)(i+1));
		for (i = 0; i < nfiles; i++) delete stfile[i].name;
        delete [] stfile; stfile = 0;
		nfiles=0;
        DeleteCriticalSection(&csDatafileRequest);
    }
	for (i = 0; i < ndatafiles; i++) {
		CLIB_FCLOSE(datafiles[i]);
        datafiles[i] = 0;
    }
}

EXTERN_C int dtf_testfile(FILE *_fp) {

    DWORD fidx = ((DWORD)_fp)-1;
    if (stfile == 0 || fidx >= MAX_OPEN_FILES) return 666;

    ST_FILE *stf = &ofile[fidx];
    EnterCriticalSection(&csDatafileRequest);

    if (CLIB_FSEEK(stf->dataf,stf->blocks[0].offset,SEEK_SET) != 0) {
        LeaveCriticalSection(&csDatafileRequest);
        return 1;
    }
    for (int block = 0; block < stf->nblocks; block++) {
        unsigned blen = stf->blocks[block].length;
        BYTE tmpBuffer[BLOCK_LEN];
        if (blen <= 0) blen = BLOCK_LEN;
        if (CLIB_FREAD(tmpBuffer,blen,1,stf->dataf) != 1) {
            LeaveCriticalSection(&csDatafileRequest);
            return 1;
        }
        xorBuf(tmpBuffer,blen,0);
        if (adler32(stf->blocks[block].offset, tmpBuffer, blen)
            != stf->blocks[block].cksum) return 2;
    }
    LeaveCriticalSection(&csDatafileRequest);
    return 0;
}

