// ***************************************************************
// MANAGER.CPP		CLIB I/O datafile interceptor			  v2.0
//
// (C) Copyright 1997-2000 7Gods demo group http://7gods.sk
// 
// Datafile manager
// ***************************************************************
#include "stdafx.h"

// ***************************************************************
typedef int (* COMPRESS_BLOCK_FUNC)
	( const lzo_byte *src, lzo_uint  src_len,
	  lzo_byte *dst, lzo_uint *dst_len,
	  lzo_voidp wrkmem );

COMPRESS_BLOCK_FUNC compress_block;
typedef void (* PROCESS_FUNC)(const char *fileName);

// ***************************************************************
char opt_chars[] = "/-";
// ***************************************************************
// extern functions
extern FLAG isOption(char *opt,int argc,char *argv[],FLAG casesensitive);
extern char *nextParam(int argc,char *argv[],int start = -1);
extern DWORD adler32(DWORD adler, const BYTE *buf, unsigned len);

// ***************************************************************
// work memory required for compression
static lzo_byte _wrkmem [ LZO1X_999_MEM_COMPRESS + 16 ];
static lzo_byte *wrkmem;

// ***************************************************************
// global variables
static ST_STOREDFILE *stfile = 0;
static FILE *dataf=NULL;		// opened datafile
static DWORD taboffs;			// file talble offset
static DWORD header[4];

// ***************************************************************
// static global variables
static unsigned allsize = 0,allcomp = 0, allnames = 0;
static DWORD nfiles=0;           // nomber of files
char datafilename[_MAX_PATH];

static FLAG recursedir = 0;

// ***************************************************************
// local function prototypes
int opendatafile(const char *_filename);
void appendstructures(void);
void AddMaskToDatafile(char *maskname);
//void CountFiles(char *maskname);
int ProcessFiles(char *maskname, PROCESS_FUNC processFile, int nFiles = 0);
void addFileToDatafile(const char *fileName);

// ***************************************************************
int main(int argc,char *argv[]) {

    int i;

	puts("\nDataFile Manager (LZ) v1.0, (c) 1997-2000 by 7Gods");

    char *fname = nextParam(argc,argv);
    if (argc < 2 || !fname) {
		puts("Usage: DFM32.EXE [options] <datafile> [<@listfiles>] [masks]\n"
		     "Example: DFM32 -r scroll.dat @maps.lst global\\*.* gfx\\*.*\n\n"
        	 "Options:\n"
             "-X ... X = 0..fast, 1..normal, 2..best compression (default = 2)\n"
             "-l ... list files in datafile\n"
             "-r ... recurse subdirectories\n"
             "-t ... test archive files\n"
        );
		return 1;
    }
    strupr(fname);
    strcpy(datafilename,fname);

    if (isOption("l",argc,argv,0)) {
   		// list files in datafile
        if (opendatafile(datafilename) != DTF_OK) {
            printf("Bad datafile %s\n",datafilename);
            return 2;
        };
        printf("\n%-32s%8s%8s\n","FILE PATH","OFFSET","LENGTH");
        puts("------------------------------------------------");
        for (i = 0; i < (int)nfiles; i++)
            printf("%-32s%8lx%8lx\n",stfile[i].name,stfile[i].offset,stfile[i].length);
        return 0;
    }

    if (isOption("t",argc,argv,0)) {
   		// test files in datafile
        if (opendatafile(datafilename) != DTF_OK) {
            printf("Bad datafile %s\n",datafilename);
            return 2;
        };
        printf("\n%-48s%8s\n","FILE PATH","STATUS");
        puts("------------------------------------------------------------");
        int errors = 0;
        for (i=0; i<(int)nfiles; i++) {
        	BYTE tmpBuffer[32768];
            FLAG error = 0;
		    DWORD nblocks = SIZE2BLOCKS(stfile[i].length);

            printf("%-48s ",stfile[i].name);

		    CBLOCK *blocks = new CBLOCK[nblocks];
            if (!blocks) { errors++; puts("bad structures."); continue; }

            fseek(dataf,stfile[i].offset,SEEK_SET);
			if (fread( blocks, sizeof(CBLOCK), nblocks, dataf)
				!= nblocks) { errors++; puts("bad structures."); continue; }

        	for (DWORD b = 0; b < nblocks; b++) {
				if (fseek(dataf,blocks[b].offset,SEEK_SET) != 0)
                	{ errors++; puts("bad structures."); error = 1; break; }
		        unsigned blen = blocks[b].length;
		        if (blen == 0) blen = 32768;
			    if (fread(tmpBuffer,blen,1,dataf) != 1) { errors++; error = 1; puts("bad structures."); break; }
	            xorBuf(tmpBuffer,blen,0);
				if (adler32(blocks[b].offset, tmpBuffer, blen)
           			!= blocks[b].cksum) { errors++; error = 1; puts("Checksum error"); break; }
            }
			delete [] blocks;
            if (!error) printf("OK.\n");
    	}
        if (errors) printf("\n!!! %d damaged files.\n",errors);
        else printf("\nAll OK.\n");
        return 0;
    }

    if (isOption("r",argc,argv,0)) recursedir = 1;
    if (isOption("0",argc,argv,0)) {
		puts("Using fast compression.\n");
		compress_block = lzo1x_1_15_compress;
    }
    else if (isOption("1",argc,argv,0)) {
		puts("Using normal compression.\n");
		compress_block = lzo1x_1_compress;
    }
    else {
		puts("Using best compression.\n");
		compress_block = lzo1x_999_compress;
    }

	wrkmem = LZO_ALIGN(_wrkmem,16);
	if (lzo_init() != LZO_E_OK) {
		printf("Initialization failed\n");
		return 666;
	}

	// make datafile
	if (strcmp(datafilename,"NUL") != 0 && (dataf = fopen(datafilename,"rb")) != NULL) {
        fclose(dataf);
    	printf("Overwrite existing %s ? (y/n)\n",datafilename);
		while (_kbhit() == 0);
		char tmp = (char)_getch();
        if (tmp != 'y' && tmp != 'Y') return 0;
		DeleteFile(datafilename);
    }

    // count total nomber of files
    nfiles = 0;
    for (fname = nextParam(argc,argv); fname != 0; fname = nextParam(argc,argv))
    	if (fname[0] == '@') {
        	char tmpname[256];
            FILE *list = fopen(fname+1,"r");
            if (list == 0) continue;
            while (fscanf(list,"%s",tmpname) == 1) nfiles += ProcessFiles(strupr(tmpname),0,nfiles);
            fclose(list);
        }
    	else nfiles += ProcessFiles(strupr(fname),0,nfiles);

	// create datafile
	if ((dataf=fopen(datafilename,"wb")) == NULL) {
    	printf("Cannot create %s",datafilename);
        return 3;
    }
	// write timestamp
    time((time_t *)header);
    fwrite(header,4,4,dataf);

    if (nfiles != 0) {
		stfile = new ST_STOREDFILE[nfiles];
    	if (stfile == 0) return DTF_NOMEMORY;

	    nfiles = 0;
	    fname = nextParam(argc,argv,1);	// datafile name
	    for (fname = nextParam(argc,argv); fname != 0; fname = nextParam(argc,argv))
           	if (fname[0] == '@') {
        		char tmpname[256];
            	FILE *list = fopen(fname+1,"r");
	            if (list == 0) continue;
    	        while (fscanf(list,"%s",tmpname) == 1) ProcessFiles(strupr(tmpname),addFileToDatafile);
            	fclose(list);
	        }
			else ProcessFiles(strupr(fname),addFileToDatafile);
    }

    printf("UnCompressed size: %d kB Compressed size: %d kB\n",allsize/1024,allcomp/1024);
    printf("Overall compression ratio ... %.1lf %%\n",100.0*allcomp/allsize);
    printf("Appending structures ... ");
    appendstructures();
    printf("OK\n");
    printf("%d files => %d(paths) + %d(data) = %d filetable\n",
		nfiles,allnames,nfiles*sizeof(ST_STOREDFILE),
		allnames + nfiles*sizeof(ST_STOREDFILE));

    delete [] stfile;
    fcloseall();
	return 0;
}

static BYTE inBuffer[BLOCK_LEN],outBuffer[BLOCK_LEN+BLOCK_LEN/64+16+3];
static unsigned int inBufLen,outBufLen,compsize;

// ***************************************************************
inline DWORD fileLength(FILE *f) {
	
	DWORD curPos = ftell(f);
	fseek(f,0,SEEK_END);
	DWORD len = ftell(f);
	fseek(f,curPos,SEEK_SET);
	return len;
}

// ***************************************************************
void addFileToDatafile(const char *fileName) {
	
	printf("Adding %-44.44s ... ",fileName);
	fflush(stdout);
	
	FILE *fromf;
	if ((fromf = fopen(fileName,"rb"))==NULL) {
		printf("ERROR\n");
		return;
	}

	DWORD fromLen = fileLength(fromf);
	// reading a file block by block with length of BLOCK_LEN
	// compress each block a write it
	DWORD nblocks = SIZE2BLOCKS(fromLen);
	if (nblocks > 65535) {
		printf("file too big\n");
		return;
	}
	CBLOCK *tmpBlocks = new CBLOCK[nblocks];
	// first reserve space for writing blocks information
	DWORD bioffset = ftell(dataf);
	bioffset = ftell(dataf);
	fseek(dataf,nblocks*sizeof(CBLOCK),SEEK_CUR);
	// comprimation itself
	DWORD nb = 0, pbin = 0;
	compsize = 0;
	while ( (inBufLen = fread(inBuffer,1,BLOCK_LEN,fromf)) > 0) {
		printf("%02d%% done\b\b\b\b\b\b\b\b",nb*100/nblocks); fflush(stdout);
		if (compress_block(inBuffer,inBufLen,
			outBuffer,&outBufLen,wrkmem) != LZO_E_OK) {
			printf("Failed ");
			break;
		}
		tmpBlocks[nb].offset = ftell(dataf);
		DWORD cksum;
		// if length of compressed block exceeds (oroginal) 32k, do NOT compress
		if (outBufLen >= BLOCK_LEN) {
			tmpBlocks[nb].length = 0;
			cksum = adler32(tmpBlocks[nb].offset, inBuffer, inBufLen);
			xorBuf(inBuffer,inBufLen,0);
			fwrite(inBuffer,inBufLen,1,dataf);
			compsize += inBufLen;
		}
		else {
			tmpBlocks[nb].length = (WORD) outBufLen;
			cksum = adler32(tmpBlocks[nb].offset, outBuffer, outBufLen);
			xorBuf(outBuffer,outBufLen,0);
			fwrite(outBuffer,outBufLen,1,dataf);
			compsize += outBufLen;
		}
		tmpBlocks[nb].cksum = cksum;
		nb++;
	}
	if (compsize >= fromLen) printf("stored  \n");
	else printf("%2d%%     \n",100lu * compsize / fromLen);
	allsize += fromLen;
	allcomp += compsize;
	if (nb != nblocks) {
		printf("Internal error!!! %d,%d\n",nb,nblocks);
		return;
	}
	// write compressed blocks
	fseek(dataf,bioffset,SEEK_SET);
	fwrite(tmpBlocks,sizeof(CBLOCK),nblocks,dataf);
	
	delete [] tmpBlocks;
	fseek(dataf,0,SEEK_END);
	
	stfile[nfiles].name = new char[strlen(fileName)+1];
	strcpy(stfile[nfiles].name,fileName);
	stfile[nfiles].offset = bioffset;
	stfile[nfiles].length = fromLen;
	nfiles++;
	
	fclose(fromf);
}

// ***************************************************************
int ProcessFiles(char *maskname, PROCESS_FUNC processFile, int nFiles) {

    WIN32_FIND_DATA lpfd;
	HANDLE findHandle;
	char ardrive[_MAX_DRIVE],ardir[_MAX_DIR],arname[_MAX_FNAME],arext[_MAX_EXT];

	_splitpath(maskname,ardrive,ardir,arname,arext);

	findHandle = FindFirstFile(maskname, &lpfd);
	if (findHandle != INVALID_HANDLE_VALUE) {
		do {
			if (lpfd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) continue;
			if (stricmp(lpfd.cFileName,datafilename) == 0) continue;
			if (processFile) {
				char realname[MAX_PATH];
				sprintf(realname,"%s%s%s",ardrive,ardir,lpfd.cFileName);
				processFile(realname);
			}
	        nFiles++;
		} while (FindNextFile(findHandle, &lpfd));
	    FindClose(findHandle);
	}

	if (!recursedir) return nFiles;
	
	char newmask[_MAX_PATH];
	_makepath(newmask,ardrive,ardir,arname,".*");
	
	findHandle = FindFirstFile(newmask, &lpfd);
	if (findHandle != INVALID_HANDLE_VALUE) {
		do {
			if ((lpfd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) == 0 || lpfd.cFileName[0] == '.') continue;
			
			strcpy(newmask,maskname);
			char *p = strrchr(newmask,'\\');
			if (p == 0) {
				sprintf(newmask,"%s\\%s",lpfd.cFileName,maskname);
			}
			else {
				p[1] = 0;	// cut off mask
				strcat(newmask,lpfd.cFileName);
				strcat(newmask,"\\");
				strcat(newmask,strrchr(maskname,'\\')+1);
			}
			nFiles += ProcessFiles(newmask,processFile,nFiles);
		} while (FindNextFile(findHandle, &lpfd));
		FindClose(findHandle);
	}
	return nFiles;
}

#define HEADER_SIZE 8
// ***********************************************************************
void appendstructures(void) {

    // special option ... NUL as datafile name => write filetable
    if (strcmp(datafilename,"NUL") == 0) {
       	FILE *f = fopen("filetab","wb");
        if (!f) return;
	    for (DWORD i=0; i<nfiles; i++) {
   			WORD namelen = (WORD)(strlen(stfile[i].name) + 1);
       		fwrite(&namelen,1,2,f);
       		fwrite(stfile[i].name,1,namelen,f);
   		    fwrite(&stfile[i].length,1,8,f);
	    }
        fclose(f);
    }

   	DWORD wheretable = ftell(dataf);
    DWORD initxor = wheretable;

    for (DWORD i=0; i<nfiles; i++) {
		WORD namelen = (WORD)(strlen(stfile[i].name) + 1);
  		fwrite(&namelen,1,2,dataf);
		allnames += namelen;
	    xorBuf(stfile[i].name,namelen,initxor);
   		fwrite(stfile[i].name,1,namelen,dataf);
		xorBuf(&stfile[i].length,8,initxor);
   		fwrite(&stfile[i].length,1,8,dataf);
   		initxor *= 0x17BC3;
	}
	fseek(dataf,0,SEEK_SET);
    header[1] = nfiles;
	header[2] = wheretable;
	header[3] = adler32(0,(BYTE *)header,12);
	fwrite(header,4,4,dataf);
}

// ***************************************************************
static int _init_error(int error) {
    fclose(dataf);
    return error;
}

// ***************************************************************
int opendatafile(const char *_filename) {

    DWORD header[4];
    ST_STOREDFILE fdesc;

    if ((dataf=fopen(_filename,"rb"))==NULL) return DTF_FILENOTFOUND;
    if (fread(header,4,4,dataf) != 4) return DTF_BADFILE;
    if (header[3] != adler32(0,(BYTE *)header,12)) return DTF_BADFILE;
    nfiles = header[1]; taboffs = header[2];

	stfile = new ST_STOREDFILE[nfiles];
    if (stfile == 0) return DTF_NOMEMORY;

    fseek(dataf,taboffs,SEEK_SET);
    DWORD initxor = taboffs;
    for (DWORD i=0; i<nfiles; i++) {
        WORD namelen;
		if (fread(&namelen,2,1,dataf)!=1) return _init_error(DTF_BADFILE);
        stfile[i].name = new char[namelen];
        if (fread(stfile[i].name,namelen,1,dataf)!=1) return _init_error(DTF_BADFILE);
        xorBuf(stfile[i].name,namelen,initxor);
        if (fread(&stfile[i].length,8,1,dataf)!=1) return _init_error(DTF_BADFILE);
        xorBuf(&stfile[i].length,8,initxor);
        initxor *= 0x17BC3;
    }
    return DTF_OK;
};
